@@ -7,14 +7,16 @@ use std::sync::Mutex;
 
 use anyhow::Context;
 use bincode::deserialize;
+use blstrs::Scalar as Fr;
 use fdlimit::raise_fd_limit;
+use ff::PrimeField;
 use filecoin_hashers::{poseidon::PoseidonHasher, Domain, HashFunction, Hasher, PoseidonArity};
 use generic_array::typenum::{Unsigned, U0, U11, U2, U8};
 use lazy_static::lazy_static;
 use log::{error, info, trace, warn};
 use merkletree::{
     merkle::{get_merkle_tree_len, is_merkle_tree_size_valid},
-    store::{Store, StoreConfig},
+    store::{DiskStore, Store, StoreConfig},
 };
 use rayon::prelude::{
     IndexedParallelIterator, IntoParallelIterator, ParallelIterator, ParallelSliceMut,
@@ -36,7 +38,7 @@ use storage_proofs_core::{
 use yastl::Pool;
 
 use crate::{
-    encode::{decode, encode},
+    encode::{decode, encode, encode_fr},
     stacked::vanilla::{
         challenges::LayerChallenges,
         column::Column,
@@ -77,6 +79,20 @@ pub struct LayerState {
     pub generated: bool,
 }
 
+pub enum TreeRElementData<Tree: MerkleTreeTrait> {
+    FrList(Vec<Fr>),
+    ElementList(Vec<<Tree::Hasher as Hasher>::Domain>),
+}
+
+#[allow(type_alias_bounds)]
+pub type PrepareTreeRDataCallback<Tree: 'static + MerkleTreeTrait> =
+    fn(
+        source: &DiskStore<<Tree::Hasher as Hasher>::Domain>,
+        data: Option<&mut Data<'_>>,
+        start: usize,
+        end: usize,
+    ) -> Result<TreeRElementData<Tree>>;
+
 impl<'a, Tree: 'static + MerkleTreeTrait, G: 'static + Hasher> StackedDrg<'a, Tree, G> {
     #[allow(clippy::too_many_arguments)]
     pub(crate) fn prove_layers(
@@ -398,7 +414,11 @@ impl<'a, Tree: 'static + MerkleTreeTrait, G: 'static + Hasher> StackedDrg<'a, Tr
         }
     }
 
-    fn build_binary_tree<K: Hasher>(
+    // NOTE: Unlike
+    // storage_proofs_core::merkle::create_base_merkle_tree, this
+    // method requires the data on disk to be exactly the same size as
+    // the tree length / NODE_SIZE.
+    pub fn build_binary_tree<K: Hasher>(
         tree_data: &[u8],
         config: StoreConfig,
     ) -> Result<BinaryMerkleTree<K>> {
@@ -417,6 +437,20 @@ impl<'a, Tree: 'static + MerkleTreeTrait, G: 'static + Hasher> StackedDrg<'a, Tr
         Ok(tree)
     }
 
+    // Even if the column builder is enabled, the GPU column builder
+    // only supports Poseidon hashes.
+    pub fn use_gpu_column_builder() -> bool {
+        SETTINGS.use_gpu_column_builder
+            && TypeId::of::<Tree::Hasher>() == TypeId::of::<PoseidonHasher>()
+    }
+
+    // Even if the tree builder is enabled, the GPU tree builder
+    // only supports Poseidon hashes.
+    pub fn use_gpu_tree_builder() -> bool {
+        SETTINGS.use_gpu_tree_builder
+            && TypeId::of::<Tree::Hasher>() == TypeId::of::<PoseidonHasher>()
+    }
+
     #[cfg(any(feature = "cuda", feature = "opencl"))]
     fn generate_tree_c<ColumnArity, TreeArity>(
         layers: usize,
@@ -429,9 +463,7 @@ impl<'a, Tree: 'static + MerkleTreeTrait, G: 'static + Hasher> StackedDrg<'a, Tr
         ColumnArity: 'static + PoseidonArity,
         TreeArity: PoseidonArity,
     {
-        if SETTINGS.use_gpu_column_builder
-            && TypeId::of::<Tree::Hasher>() == TypeId::of::<PoseidonHasher>()
-        {
+        if Self::use_gpu_column_builder() {
             Self::generate_tree_c_gpu::<ColumnArity, TreeArity>(
                 layers,
                 nodes_count,
@@ -488,10 +520,8 @@ impl<'a, Tree: 'static + MerkleTreeTrait, G: 'static + Hasher> StackedDrg<'a, Tr
         use std::sync::mpsc::sync_channel as channel;
         use std::sync::{Arc, RwLock};
 
-        use blstrs::Scalar as Fr;
         use fr32::fr_into_bytes;
         use generic_array::GenericArray;
-        use merkletree::store::DiskStore;
         use neptune::{
             batch_hasher::Batcher,
             column_tree_builder::{ColumnTreeBuilder, ColumnTreeBuilderTrait},
@@ -803,29 +833,118 @@ impl<'a, Tree: 'static + MerkleTreeTrait, G: 'static + Hasher> StackedDrg<'a, Tr
         })
     }
 
+    fn prepare_tree_r_data_cpu(
+        source: &DiskStore<<Tree::Hasher as Hasher>::Domain>,
+        data: Option<&mut Data<'_>>,
+        start: usize,
+        end: usize,
+    ) -> Result<TreeRElementData<Tree>> {
+        let encoded_data: Vec<<Tree::Hasher as Hasher>::Domain> = source
+            .read_range(start..end)?
+            .into_par_iter()
+            .zip(
+                data.expect("failed to unwrap data").as_mut()
+                    [(start * NODE_SIZE)..(end * NODE_SIZE)]
+                    .par_chunks_mut(NODE_SIZE),
+            )
+            .map(|(key, data_node_bytes)| {
+                let data_node = <Tree::Hasher as Hasher>::Domain::try_from_bytes(data_node_bytes)
+                    .expect("try from bytes failed");
+
+                let key_elem = <Tree::Hasher as Hasher>::Domain::try_from_bytes(&key.into_bytes())
+                    .expect("failed to convert key");
+                let encoded_node = encode::<<Tree::Hasher as Hasher>::Domain>(key_elem, data_node);
+                data_node_bytes.copy_from_slice(AsRef::<[u8]>::as_ref(&encoded_node));
+
+                encoded_node
+            })
+            .collect();
+
+        Ok(TreeRElementData::ElementList(encoded_data))
+    }
+
     #[cfg(any(feature = "cuda", feature = "opencl"))]
-    fn generate_tree_r_last<TreeArity>(
+    fn prepare_tree_r_data(
+        source: &DiskStore<<Tree::Hasher as Hasher>::Domain>,
+        data: Option<&mut Data<'_>>,
+        start: usize,
+        end: usize,
+    ) -> Result<TreeRElementData<Tree>> {
+        if Self::use_gpu_tree_builder() {
+            use fr32::bytes_into_fr;
+
+            let mut layer_bytes = vec![0u8; (end - start) * std::mem::size_of::<Fr>()];
+            source
+                .read_range_into(start, end, &mut layer_bytes)
+                .expect("failed to read layer bytes");
+
+            let encoded_data: Vec<_> = layer_bytes
+                .into_par_iter()
+                .chunks(std::mem::size_of::<Fr>())
+                .map(|chunk| bytes_into_fr(&chunk).expect("Could not create Fr from bytes."))
+                .zip(
+                    data.expect("failed to unwrap data").as_mut()
+                        [(start * NODE_SIZE)..(end * NODE_SIZE)]
+                        .par_chunks_mut(NODE_SIZE),
+                )
+                .map(|(key, data_node_bytes)| {
+                    let data_node =
+                        <Tree::Hasher as Hasher>::Domain::try_from_bytes(data_node_bytes)
+                            .expect("try_from_bytes failed");
+
+                    let mut encoded_fr: Fr = key;
+                    let data_node_fr: Fr = data_node.into();
+                    encode_fr(&mut encoded_fr, data_node_fr);
+                    let encoded_fr_repr = encoded_fr.to_repr();
+                    data_node_bytes.copy_from_slice(AsRef::<[u8]>::as_ref(&encoded_fr_repr));
+
+                    encoded_fr
+                })
+                .collect();
+
+            Ok(TreeRElementData::FrList(encoded_data))
+        } else {
+            Self::prepare_tree_r_data_cpu(source, data, start, end)
+        }
+    }
+
+    #[cfg(not(any(feature = "cuda", feature = "opencl")))]
+    fn prepare_tree_r_data(
+        source: &DiskStore<<Tree::Hasher as Hasher>::Domain>,
+        data: Option<&mut Data<'_>>,
+        start: usize,
+        end: usize,
+    ) -> Result<TreeRElementData<Tree>> {
+        Self::prepare_tree_r_data_cpu(source, data, start, end)
+    }
+
+    #[cfg(any(feature = "cuda", feature = "opencl"))]
+    pub fn generate_tree_r_last<TreeArity>(
         data: &mut Data<'_>,
         nodes_count: usize,
         tree_count: usize,
         tree_r_last_config: StoreConfig,
         replica_path: PathBuf,
-        labels: &LabelsCache<Tree>,
+        source: &DiskStore<<Tree::Hasher as Hasher>::Domain>,
+        callback: Option<PrepareTreeRDataCallback<Tree>>,
     ) -> Result<LCTree<Tree::Hasher, Tree::Arity, Tree::SubTreeArity, Tree::TopTreeArity>>
     where
         TreeArity: PoseidonArity,
     {
-        // The GPU tree builder only support Poseidon hashes.
-        if SETTINGS.use_gpu_tree_builder
-            && TypeId::of::<Tree::Hasher>() == TypeId::of::<PoseidonHasher>()
-        {
+        let encode_data = match callback {
+            Some(x) => x,
+            None => Self::prepare_tree_r_data,
+        };
+
+        if Self::use_gpu_tree_builder() {
             Self::generate_tree_r_last_gpu::<TreeArity>(
                 data,
                 nodes_count,
                 tree_count,
                 tree_r_last_config,
                 replica_path,
-                labels,
+                source,
+                encode_data,
             )
         } else {
             Self::generate_tree_r_last_cpu::<TreeArity>(
@@ -834,30 +953,38 @@ impl<'a, Tree: 'static + MerkleTreeTrait, G: 'static + Hasher> StackedDrg<'a, Tr
                 tree_count,
                 tree_r_last_config,
                 replica_path,
-                labels,
+                source,
+                encode_data,
             )
         }
     }
 
     #[cfg(not(any(feature = "cuda", feature = "opencl")))]
-    fn generate_tree_r_last<TreeArity>(
+    pub fn generate_tree_r_last<TreeArity>(
         data: &mut Data<'_>,
         nodes_count: usize,
         tree_count: usize,
         tree_r_last_config: StoreConfig,
         replica_path: PathBuf,
-        labels: &LabelsCache<Tree>,
+        source: &DiskStore<<Tree::Hasher as Hasher>::Domain>,
+        callback: Option<PrepareTreeRDataCallback<Tree>>,
     ) -> Result<LCTree<Tree::Hasher, Tree::Arity, Tree::SubTreeArity, Tree::TopTreeArity>>
     where
         TreeArity: PoseidonArity,
     {
+        let encode_data = match callback {
+            Some(x) => x,
+            None => Self::prepare_tree_r_data,
+        };
+
         Self::generate_tree_r_last_cpu::<TreeArity>(
             data,
             nodes_count,
             tree_count,
             tree_r_last_config,
             replica_path,
-            labels,
+            source,
+            encode_data,
         )
     }
 
@@ -868,7 +995,8 @@ impl<'a, Tree: 'static + MerkleTreeTrait, G: 'static + Hasher> StackedDrg<'a, Tr
         tree_count: usize,
         tree_r_last_config: StoreConfig,
         replica_path: PathBuf,
-        labels: &LabelsCache<Tree>,
+        source: &DiskStore<<Tree::Hasher as Hasher>::Domain>,
+        callback: PrepareTreeRDataCallback<Tree>,
     ) -> Result<LCTree<Tree::Hasher, Tree::Arity, Tree::SubTreeArity, Tree::TopTreeArity>>
     where
         TreeArity: PoseidonArity,
@@ -878,7 +1006,6 @@ impl<'a, Tree: 'static + MerkleTreeTrait, G: 'static + Hasher> StackedDrg<'a, Tr
         use std::io::Write;
         use std::sync::mpsc::sync_channel as channel;
 
-        use blstrs::Scalar as Fr;
         use fr32::fr_into_bytes;
         use merkletree::merkle::{get_merkle_tree_cache_size, get_merkle_tree_leafs};
         use neptune::{
@@ -893,9 +1020,6 @@ impl<'a, Tree: 'static + MerkleTreeTrait, G: 'static + Hasher> StackedDrg<'a, Tr
             tree_count,
         )?;
 
-        data.ensure_data()?;
-        let last_layer_labels = labels.labels_for_last_layer()?;
-
         info!("generating tree r last using the GPU");
         let max_gpu_tree_batch_size = SETTINGS.max_gpu_tree_batch_size as usize;
 
@@ -928,44 +1052,14 @@ impl<'a, Tree: 'static + MerkleTreeTrait, G: 'static + Hasher> StackedDrg<'a, Tr
                             end,
                         );
 
-                        let encoded_data = {
-                            use fr32::bytes_into_fr;
-
-                            let mut layer_bytes =
-                                vec![0u8; (end - start) * std::mem::size_of::<Fr>()];
-                            last_layer_labels
-                                .read_range_into(start, end, &mut layer_bytes)
-                                .expect("failed to read layer bytes");
-
-                            layer_bytes
-                                .into_par_iter()
-                                .chunks(std::mem::size_of::<Fr>())
-                                .map(|chunk| {
-                                    bytes_into_fr(&chunk).expect("Could not create Fr from bytes.")
-                                })
-                                .zip(
-                                    data.as_mut()[(start * NODE_SIZE)..(end * NODE_SIZE)]
-                                        .par_chunks_mut(NODE_SIZE),
-                                )
-                                .map(|(key, data_node_bytes)| {
-                                    let data_node =
-                                        <Tree::Hasher as Hasher>::Domain::try_from_bytes(
-                                            data_node_bytes,
-                                        )
-                                        .expect("try_from_bytes failed");
-
-                                    let encoded_node = encode::<<Tree::Hasher as Hasher>::Domain>(
-                                        key.into(),
-                                        data_node,
-                                    );
-                                    data_node_bytes
-                                        .copy_from_slice(AsRef::<[u8]>::as_ref(&encoded_node));
-
-                                    encoded_node
-                                })
+                        let prepared_data = match callback(source, Some(data), start, end)
+                            .expect("failed to prepare tree_r_last data")
+                        {
+                            TreeRElementData::FrList(x) => x,
+                            _ => panic!("fr_list is required"),
                         };
-
                         node_index += chunked_nodes_count;
+
                         trace!(
                             "node index {}/{}/{}",
                             node_index,
@@ -973,13 +1067,10 @@ impl<'a, Tree: 'static + MerkleTreeTrait, G: 'static + Hasher> StackedDrg<'a, Tr
                             nodes_count,
                         );
 
-                        let encoded: Vec<_> =
-                            encoded_data.into_par_iter().map(|x| x.into()).collect();
-
                         let is_final = node_index == nodes_count;
                         builder_tx
-                            .send((encoded, is_final))
-                            .expect("failed to send encoded");
+                            .send((prepared_data, is_final))
+                            .expect("failed to send prepared data");
                     }
                 }
             });
@@ -1002,13 +1093,13 @@ impl<'a, Tree: 'static + MerkleTreeTrait, G: 'static + Hasher> StackedDrg<'a, Tr
                 // Loop until all trees for all configs have been built.
                 for i in 0..config_count {
                     loop {
-                        let (encoded, is_final) =
-                            builder_rx.recv().expect("failed to recv encoded data");
+                        let (prepared_data, is_final) =
+                            builder_rx.recv().expect("failed to recv prepared data");
 
                         // Just add non-final leaf batches.
                         if !is_final {
                             tree_builder
-                                .add_leaves(&encoded)
+                                .add_leaves(&prepared_data)
                                 .expect("failed to add leaves");
                             continue;
                         };
@@ -1020,7 +1111,7 @@ impl<'a, Tree: 'static + MerkleTreeTrait, G: 'static + Hasher> StackedDrg<'a, Tr
                             tree_count
                         );
                         let (_, tree_data) = tree_builder
-                            .add_final_leaves(&encoded)
+                            .add_final_leaves(&prepared_data)
                             .expect("failed to add final leaves");
 
                         writer_tx.send(tree_data).expect("failed to send tree_data");
@@ -1083,7 +1174,8 @@ impl<'a, Tree: 'static + MerkleTreeTrait, G: 'static + Hasher> StackedDrg<'a, Tr
         tree_count: usize,
         tree_r_last_config: StoreConfig,
         replica_path: PathBuf,
-        labels: &LabelsCache<Tree>,
+        source: &DiskStore<<Tree::Hasher as Hasher>::Domain>,
+        callback: PrepareTreeRDataCallback<Tree>,
     ) -> Result<LCTree<Tree::Hasher, Tree::Arity, Tree::SubTreeArity, Tree::TopTreeArity>>
     where
         TreeArity: PoseidonArity,
@@ -1095,31 +1187,20 @@ impl<'a, Tree: 'static + MerkleTreeTrait, G: 'static + Hasher> StackedDrg<'a, Tr
             tree_count,
         )?;
 
-        data.ensure_data()?;
-        let last_layer_labels = labels.labels_for_last_layer()?;
-
         info!("generating tree r last using the CPU");
-        let size = Store::len(last_layer_labels);
 
+        // Note that nodes_count is the count of nodes in each base tree
         let mut start = 0;
-        let mut end = size / tree_count;
+        let mut end = nodes_count;
 
         for (i, config) in configs.iter().enumerate() {
-            let encoded_data = last_layer_labels
-                .read_range(start..end)?
-                .into_par_iter()
-                .zip(
-                    data.as_mut()[(start * NODE_SIZE)..(end * NODE_SIZE)].par_chunks_mut(NODE_SIZE),
-                )
-                .map(|(key, data_node_bytes)| {
-                    let data_node =
-                        <Tree::Hasher as Hasher>::Domain::try_from_bytes(data_node_bytes)
-                            .expect("try from bytes failed");
-                    let encoded_node = encode::<<Tree::Hasher as Hasher>::Domain>(key, data_node);
-                    data_node_bytes.copy_from_slice(AsRef::<[u8]>::as_ref(&encoded_node));
-
-                    encoded_node
-                });
+            let encoded_data: Vec<<Tree::Hasher as Hasher>::Domain> =
+                match callback(source, Some(data), start, end)
+                    .expect("failed to prepare tree_r_last data")
+                {
+                    TreeRElementData::ElementList(x) => x,
+                    _ => panic!("element list required"),
+                };
 
             info!(
                 "building base tree_r_last with CPU {}/{}",
@@ -1146,8 +1227,8 @@ impl<'a, Tree: 'static + MerkleTreeTrait, G: 'static + Hasher> StackedDrg<'a, Tr
             )
             .with_context(|| format!("failed tree_r_last CPU {}/{}", i + 1, tree_count))?;
 
-            start = end;
-            end += size / tree_count;
+            start += nodes_count;
+            end += nodes_count;
         }
 
         create_lc_tree::<LCTree<Tree::Hasher, Tree::Arity, Tree::SubTreeArity, Tree::TopTreeArity>>(
@@ -1329,6 +1410,9 @@ impl<'a, Tree: 'static + MerkleTreeTrait, G: 'static + Hasher> StackedDrg<'a, Tr
         drop(tree_d);
 
         // Encode original data into the last layer.
+        let last_layer_labels = labels.labels_for_last_layer()?;
+        data.ensure_data()?;
+
         info!("building tree_r_last");
         let tree_r_last = measure_op(Operation::GenerateTreeRLast, || {
             Self::generate_tree_r_last::<Tree::Arity>(
@@ -1337,7 +1421,8 @@ impl<'a, Tree: 'static + MerkleTreeTrait, G: 'static + Hasher> StackedDrg<'a, Tr
                 tree_count,
                 tree_r_last_config.clone(),
                 replica_path.clone(),
-                &labels,
+                last_layer_labels,
+                None,
             )
             .context("failed to generate tree_r_last")
         })?;
@@ -1431,7 +1516,6 @@ impl<'a, Tree: 'static + MerkleTreeTrait, G: 'static + Hasher> StackedDrg<'a, Tr
         use std::fs::OpenOptions;
         use std::io::Write;
 
-        use blstrs::Scalar as Fr;
         use ff::Field;
         use fr32::fr_into_bytes;
         use merkletree::merkle::{get_merkle_tree_cache_size, get_merkle_tree_leafs};
@@ -1447,7 +1531,7 @@ impl<'a, Tree: 'static + MerkleTreeTrait, G: 'static + Hasher> StackedDrg<'a, Tr
             tree_count,
         )?;
 
-        if SETTINGS.use_gpu_tree_builder {
+        if Self::use_gpu_tree_builder() {
             info!("generating tree r last using the GPU");
             let max_gpu_tree_batch_size = SETTINGS.max_gpu_tree_batch_size as usize;
 
